
// Ignore warnings as these are auto-generated headers made by a third party
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
#pragma GCC diagnostic ignored "-Waddress-of-packed-member"
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
#include <mavlink_lib/pyxis/mavlink.h>
#pragma GCC diagnostic pop

#include <fstream>
#include <iostream>
#include <string>
#include <unordered_map>

const mavlink_message_info_t &get_message_info_for(uint8_t msgid)
{
    static const mavlink_message_info_t infos[256] = MAVLINK_MESSAGE_INFO;
    return infos[msgid];
}

class CsvStream
{
public:
    CsvStream(std::string path, const mavlink_message_info_t &info)
        : os(), info(info)
    {

        os.exceptions(std::ifstream::failbit);
        os.open(path, std::ios::binary);

        format_message_header();
    }

    void put(const mavlink_message_t &msg) { format_message(msg); }

private:
    void format_field(const mavlink_message_t &msg, int id)
    {
        const mavlink_field_info_t &field = info.fields[id];
        if (field.array_length != 0)
        {
            os << "<array not implemented>";
        }
        else
        {
            switch (field.type)
            {
                case MAVLINK_TYPE_CHAR:
                    os << _MAV_RETURN_char(&msg, field.wire_offset);
                    break;

                case MAVLINK_TYPE_UINT8_T:
                    os << static_cast<uint64_t>(
                        _MAV_RETURN_uint8_t(&msg, field.wire_offset));
                    break;

                case MAVLINK_TYPE_UINT16_T:
                    os << static_cast<uint64_t>(
                        _MAV_RETURN_uint16_t(&msg, field.wire_offset));
                    break;

                case MAVLINK_TYPE_UINT32_T:
                    os << static_cast<uint64_t>(
                        _MAV_RETURN_uint32_t(&msg, field.wire_offset));
                    break;

                case MAVLINK_TYPE_UINT64_T:
                    os << static_cast<uint64_t>(
                        _MAV_RETURN_uint64_t(&msg, field.wire_offset));
                    break;

                case MAVLINK_TYPE_INT8_T:
                    os << static_cast<int64_t>(
                        _MAV_RETURN_uint8_t(&msg, field.wire_offset));
                    break;

                case MAVLINK_TYPE_INT16_T:
                    os << static_cast<int64_t>(
                        _MAV_RETURN_uint16_t(&msg, field.wire_offset));
                    break;

                case MAVLINK_TYPE_INT32_T:
                    os << static_cast<int64_t>(
                        _MAV_RETURN_uint32_t(&msg, field.wire_offset));
                    break;

                case MAVLINK_TYPE_INT64_T:
                    os << static_cast<int64_t>(
                        _MAV_RETURN_uint64_t(&msg, field.wire_offset));
                    break;

                case MAVLINK_TYPE_FLOAT:
                    os << _MAV_RETURN_float(&msg, field.wire_offset);
                    break;

                case MAVLINK_TYPE_DOUBLE:
                    os << _MAV_RETURN_double(&msg, field.wire_offset);
                    break;
            }
        }
    }

    void format_message(const mavlink_message_t &msg)
    {
        bool first = true;
        for (unsigned i = 0; i < info.num_fields; i++)
        {
            if (!first)
            {
                os << ",";
            }

            format_field(msg, i);
            first = false;
        }

        os << std::endl;
    }

    void format_message_header()
    {
        bool first = true;
        for (unsigned i = 0; i < info.num_fields; i++)
        {
            if (!first)
            {
                os << ",";
            }

            os << std::string(info.fields[i].name);
            first = false;
        }

        os << std::endl;
    }

    std::ofstream os;
    const mavlink_message_info_t &info;
};

class Decoder
{
public:
    Decoder(std::string input, std::string output)
        : input(input), output(output)
    {
    }

    void decode()
    {
        std::ifstream is;
        is.exceptions(std::ifstream::failbit);
        is.open(input, std::ios::binary);

        std::unordered_map<uint8_t, CsvStream> streams;

        mavlink_message_t msg;
        mavlink_status_t status;
        while (is.peek() != EOF)
        {
            if (mavlink_parse_char(MAVLINK_COMM_0, is.get(), &msg, &status))
            {
                uint8_t msgid = msg.msgid;
                const mavlink_message_info_t &info =
                    get_message_info_for(msgid);

                auto it = streams.find(msgid);
                if (it == streams.end())
                {
                    std::string filename = get_filename_for(info);
                    it = streams.insert({msgid, CsvStream(filename, info)})
                             .first;
                }

                it->second.put(msg);
            }
        }
    }

private:
    std::string get_filename_for(const mavlink_message_info_t &info)
    {
        return output + "/" + info.name + ".csv";
    }

    std::string input;
    std::string output;
};

void print_usage()
{
    std::cout << "Skyward mavlink decoder 2.0\n"
                 "\n"
                 "USAGE:\n"
                 "    ./mavdecoder <FILENAME>\n"
                 "\n"
                 "OPTIONS:\n"
                 "    FILENAME     Filename of a mavlink encoded file\n"
                 "\n"
              << std::flush;
}

int main(int argc, char *argv[])
{
    if (argc != 2)
    {
        print_usage();
        return -1;
    }

    try
    {
        Decoder decoder(argv[1], "out");
        decoder.decode();
    }
    catch (std::exception &ex)
    {
        std::cout << "An exception occurred: " << ex.what() << std::endl;
        return -1;
    }

    return 0;
}